import sys
import omc048
import obs
import scheduler
import config
import driver_manager
import logging_manager
import ftp
import usb_wake
syslog = obs.system_log('APPLICATION')

###   SYSTEM   ###
# Set hardfaults to reboot the logger
obs.fault_debug(False)
# Run board diagnostics
diagnostics = omc048.diagnostics()
diagnostics.all()
# Read the config file
configuration = config.get_config()
if configuration:
    syslog.info('config.txt detected')
else:
    sys.exit()
# Define the level of the syslogs from configuration
obs.syslog_level_repl(configuration['Omc048']['repl_log_level'])
obs.syslog_level_sd(configuration['Omc048']['file_log_level'])
    
# Inform the application is started
syslog.info('Application started: {}'.format(configuration['Omc048']['application']))
syslog.info('File Syslog level: {}'.format(configuration['Omc048']['file_log_level']))
syslog.info('Repl Syslog level: {}'.format(configuration['Omc048']['repl_log_level']))
# Obtain utc offset from the configuration
utc_offset = [config.get_config_or_default(configuration['Omc048'],'utc_time_offset_hours', 0), 
              config.get_config_or_default(configuration['Omc048'],'utc_time_offset_minutes', 0)]
omc048.RTC().utc_offset(tuple(utc_offset))
syslog.info('UTC time device offset: {}(hrs):{}(min)'.format(utc_offset[0], utc_offset[1]))

###   INITIALISATION   ###
drivers = driver_manager.SensorDrvManager(configuration)

###   PREPARATION   ###
log_mngr = logging_manager.LoggingManager(configuration)
log_mngr.add_sensor_params(drivers)

ftp_obj = ftp.Ftp(configuration, log_mngr)
usb_obj = usb_wake.usb_wake(configuration, log_mngr)

# Start logging data
log_mngr.start()

async def sensor_update(sen_group, index=0):
    try:
        await drivers.sensors[configuration[sen_group][index]['id']].get_data()
        log_mngr.add_sensor_data(configuration[sen_group][index]['id'], drivers)
    except Exception as err:
        syslog.error('sen_group: {}, index {}, config {}, {}: {}'.format(sen_group, index, configuration[sen_group], type(err).__name__, err))

def schedule_sensor(sensor_group):
    group_cfg = config.get_config_params(configuration, sensor_group)
    if not group_cfg:
        return
    for idx, sen_conf in enumerate(group_cfg):
        if config.get_config_params(drivers.sensors, config.get_config_params(sen_conf, 'id', True)):
            scheduler.schedule_task(sensor_update, sen_conf['sample_interval'], '{}_{}'.format(sensor_group, idx), [sensor_group, idx])

#Schedule tasks for every driver corresponding cron configuration
ser_keys = driver_manager.serial_drivers.keys()
for sen_group in ser_keys:
    schedule_sensor(sen_group)
non_ser_keys = driver_manager.non_serial_drivers.keys()
for sen_group in non_ser_keys:
    schedule_sensor(sen_group)

#Define pre and post sleep function executes
def pre_sleep():
    drivers.deinit()
    if not omc048.usb().is_connected():
        omc048.usb().power(False)
  
def post_sleep():
    drivers.reinit()

if config.get_config_params(configuration, 'Data_file', True):
    scheduler.schedule_task(log_mngr.create_new_file, config.get_config_params(configuration['Data_file'][0], 'create_interval', True), 'file_create')
if config.get_config_params(configuration, 'Ftp'):
    scheduler.schedule_task(ftp_obj.ftp_transmit, config.get_config_params(configuration['Ftp'][0], 'transmit_interval', True), 'file_transmit')
if config.get_config_params(configuration, 'Usb_wake'):
    scheduler.schedule_task(usb_obj.usb_check, config.get_config_params(configuration['Usb_wake'][0], 'wake_interval', True), 'usb_wake_interval')

scheduler.sleep_between_tasks(pre_sleep,post_sleep)

scheduler.start()

obs.hard_reset()
