import re
import obs
import omc048
import config
from read_line import Read_line
from sensor_drv import SensorDrv
from util import parse_int, strlist_to_intlist, lookup_parameters

class EXO(SensorDrv):
    __syslog = obs.system_log('EXO')
    
    def __init__(self, serial, conf):
        SensorDrv.__init__(self)
        self.__available_parameters =  {
                                            255:    ("Unknown Parameter" ,"??"      ,"????" ),
                                            1:      ("Temp C"	         ,"C"       ,"TMP"  ), 
                                            2:      ("Temp F"	         ,"F"       ,"TMP"  ), 
                                            3:      ("Temp K"	         ,"K"       ,"TMP"  ), 
                                            4:      ("Cond mS/cm"	     ,"mS/cm"   ,"CON"  ), 
                                            5:      ("Cond uS/cm"        ,"uS/cm"   ,"CON"  ), 
                                            6:      ("SpCond ms/cm"      ,"mS/cm"   ,"SPC"  ), 
                                            7:      ("SpCond uS/cm"      ,"uS/cm"   ,"SPC"  ), 
                                            10:     ("TDS g/L"           ,"g/L"     ,"TDS"  ), 
                                            12:     ("Sal psu"           ,"psu"     ,"SAL"  ), 
                                            17:     ("pH mV"             ,"mV"      ,"PHmV" ), 
                                            18:     ("pH"                ,"pH"      ,"PH"   ), 
                                            19:     ("Orp mV"            ,"mV"      ,"ORP"  ), 
                                            20:     ("Press psia"        ,"psi"     ,"PSIA" ), 
                                            21:     ("Press psig"        ,"psi"     ,"PSIG" ), 
                                            22:     ("Depth meters"      ,"m"       ,"DEP"  ), 
                                            23:     ("Depth feet"        ,"f"       ,"DEP"  ), 
                                            28:     ("Battery volts"     ,"V"       ,"BTV"  ), 
                                            37:     ("Turbidity"         ,"NTU"     ,"TUR"  ), 
                                            47:     ("NH3- N mg/L"       ,"mg/L"    ,"NH3"  ), 
                                            48:     ("NH4+ N mg/L"       ,"mg/L"    ,"NH4"  ),   
                                            95:     ("TDS kg/L"          ,"kg/L"    ,"TDS"  ), 
                                            101:    ("NO3- N mV"         ,"mV"     ,"NO3"   ), 
                                            106:    ("NO3- N mg/L"       ,"mg/L"   ,"NO3"   ), 
                                            108:    ("NH4+ N mV"         ,"mV"     ,"NH4"   ), 
                                            110:    ("TDS mg/L"          ,"mg/L"   ,"TDS"   ), 
                                            112:    ("Cl- mg/L"          ,"mg/L"   ,"CL"    ), 
                                            145:    ("Cl- mV"            ,"mV"     ,"CL"    ), 
                                            190:    ("TSS mg/L"          ,"mg/L"   ,"TSS"   ), 
                                            191:    ("TSS g/L"           ,"g/L"    ,"TSS"   ), 
                                            193:    ("Chlorophyll ug/L"  ,"ug/L"   ,"CHL"   ), 
                                            194:    ("Chlorophyll RFU"   ,"RFU"    ,"CHLr"  ), 
                                            211:    ("ODO % sat"         ,"%"      ,"ODO"   ), 
                                            212:    ("ODO mg/L"          ,"mg/L"   ,"ODOm"  ), 
                                            214:    ("ODO % local"       ,"%"      ,"ODO"   ), 
                                            216:    ("BGA PC RFU"        ,"RFU"    ,"BGAC"  ), 
                                            218:    ("BGA PE RFU"        ,"RFU"    ,"BGAE"  ), 
                                            223:    ("Turbidity FNU"     ,"FNU"    ,"TUR"   ), 
                                            224:    ("Turbidity RAW"     ,"RAW"    ,"TUR"   ), 
                                            225:    ("BGA PC ug/L"       ,"ug/L"   ,"BGAC"  ), 
                                            226:    ("BGA PE ug/L"       ,"ug/L"   ,"BGAE"  ), 
                                            227:    ("fDOM RFU"          ,"RFU"    ,"FDOM"  ), 
                                            228:    ("fDOM QSU"          ,"QSU"    ,"FDOM"  ), 
                                            229:    ("Wiper Position"    ,"V"      ,"WPOS"  ), 
                                            230:    ("Cable Power volts" ,"V"      ,"CPV"   ), 
                                            231:    ("BGA PC Raw"        ,"RAW"    ,"BGAC"  ), 
                                            232:    ("BGA PE Raw"        ,"RAW"    ,"BGAE"  ), 
                                            233:    ("fDOM Raw"          ,"RAW"    ,"FDOM"  ), 
                                            234:    ("Chlorophyll Raw"   ,"RAW"    ,"CHL"   ), 
                                            237:    ("nLF Cond mS/cm"    ,"mS/cm"  ,"NCON"  ), 
                                            238:    ("nLF Cond uS/cm"    ,"uS/cm"  ,"NCON"  ), 
                                            239:    ("Wiper Peak Current","mA"     ,"WPMA"  ),
                                            242:    ("Chlorofyl"         ,"cells/ml","CHLML")}
        self.startup_time = 30
        self.samples_between_wipes = config.get_config_or_default(conf, 'samples_per_wipe', 10)
        self.__syslog.info('Samples per wipe set to: {}'.format(self.samples_between_wipes))
        self.__serial_data = Read_line(serial)
        self.__serial = serial
        self.__parameter_size = 0
        self.__samples_count = self.samples_between_wipes
        self.__wipe_flag = wiper_state.IDLE

    def __check_wiper_status(self):
        # Note: After a wipe command the hwipesleft command does not immediately respond with non-zero values!
        # The first few hwipesleft commands respond with a '0', after a short duration the responses are as expected during wipe (non-zero)
        # Therefore the "INITIALIZING" state is introduced.
        self.__send_command('hwipesleft')
        response = self.__get_sensor_response()
        if self.__wipe_flag == wiper_state.INITIALIZING:
            if response and response[1] != '0\r\n':
                self.__wipe_flag = wiper_state.BUSY
            return True
        if response and self.__wipe_flag == wiper_state.BUSY and response[1] != '0\r\n':
            return True
        return False

    def __send_wipe_command(self):
        self.__syslog.info('Wipe command initiated')
        self.__send_command('twipeb')
        response = self.__get_sensor_response()
        #if there is no wiper attatched the twipeb command responds with a '0', in this case the INITIALIZING state is skipped
        if response and response[1] == '0\r\n':
            self.__syslog.warn('Wiper not detected')
            self.__wipe_flag = wiper_state.BUSY
        else:
            self.__wipe_flag = wiper_state.INITIALIZING
        self.__samples_count = 0

    def read_values(self):  
        #in case the sensor fell asleep, wake it with a dummy, before obtaining data. the response can be ignored
        self.__send_command('0')
        self.__get_sensor_response(True)
        if self.samples_between_wipes != 0:
            if self.__check_wiper_status():
                return False
            if self.__samples_count >= self.samples_between_wipes and self.__wipe_flag == wiper_state.IDLE:
                self.__send_wipe_command()
                return False
        self.__send_command('data')
        response = self.__get_sensor_response()
        if response:
            response_split = response[1].split(' ')
            response_split.remove('\r\n')
            if len(response_split) == self.__parameter_size:
                self.data = response_split
                if self.samples_between_wipes != 0:
                    self.__samples_count += 1
                    self.__wipe_flag = wiper_state.IDLE
                return True
            else:
                self.__syslog.warn('Data length does not correspond to parameters length')
        return False

    def __send_command(self, command):
        command = command + '\r'
        self.__serial.write(command)

    def __get_sensor_response(self, syslog_hide = False):
        line_count = 0
        response = []
        response_timeout = obs.millis()
        while line_count < 2:
            line_count = self.__serial_data.update()
            #in case the sensor does not respond
            if obs.elapsed_millis(response_timeout) > 5000:
                if not syslog_hide:
                    self.__syslog.warn('Sensor response timeout')
                return []

        for _ in range(line_count):
            returned_response = self.__serial_data.get()
            if type(returned_response) is bytes:
                #In case of framing errors, correct the sentence
                returned_response = re.sub(b"[\x7f-\xff]",b"", returned_response)
                returned_response = returned_response.decode()
            response.append(returned_response)
        return(response)

    def __wake_up_sensor(self):
        #In case the pwruptorun 1 was true, a 0/r must be sent to stop automatic tansmission
        self.__send_command('0')
        self.__get_sensor_response(True)
        self.__send_command('setecho 1')
        response = self.__get_sensor_response()
        if response and self.__check_response(response,'setecho 1') == True:
            self.__serial_data.flush()
            return True
        return False

    def __check_response(self, response, command):
        if response[0] == '# {}{}'.format(command, '\r\n') and response[1] == 'OK\r\n':
            return True
        return False

    def __initialise_sensor(self):
        init_commands = ("setecho 1",                              # interface echo on 
                         "pwruptorun 0",                           # no power up to run
                         "setdelim 0",                             # data fields delimited by space
                         "setmode 0",                              # set to normal mode
                         "setradix 0",                             # use point as decimal designator
                         "setsonde 0")                             # only one sonde is supported
        for cmd in init_commands:
            self.__send_command(cmd)
            response = self.__get_sensor_response()
            if response:
                if not self.__check_response(response,cmd):
                    return False
            else:
                return False
        return True

    def __retreive_parameters(self):
        self.__send_command('para')
        response = self.__get_sensor_response()
        if response:
            response_split = response[1].split(' ')
            response_int = strlist_to_intlist(response_split)
            self.__parameter_size = len(response_int)
            return response_int
        return None

    def detect_parameters(self, tag_number = ''): 
        if self.__wake_up_sensor() and self.__initialise_sensor():
            resp = self.__retreive_parameters()
            if resp:
                self.parameters = lookup_parameters(resp, self.__available_parameters, tag_number)
                return True
        return False
        
    def flush(self):
        self.__serial_data.flush()


class wiper_state:
    IDLE = 0
    BUSY = 1
    INITIALIZING = 2

