import omc048
import obs
import config
from read_line import Read_line
from nmea import parser
from nmea_decoder import NmeaDecode, VALUE, PARAM
from sensor_drv import SensorDrv
from util import parse_int, parse_float, get_datetime_with_offset

class NavilockMd6(SensorDrv):
  __syslog = obs.system_log('GPS')
  __nmea_decoder = NmeaDecode()
  
  def __init__(self, serial, conf):
    SensorDrv.__init__(self)
    self.__sensor_data_valid = False
    self.__data = {}
    self.__params = {}
    self.datetime = ()
    self.__serial_data = Read_line(serial)
    self.__time_deviation_in_sec = 2.0
    self.__utc_sync = config.get_config_or_default(conf, 'utc_time_sync', False)
    self.__utc_offset = omc048.RTC().utc_offset()
    if self.__utc_sync:
      self.__syslog.info('UTC automatic time synchronisation enabled on {}'.format(conf['id']))
 
  def read_values(self, detect_params=False):
    line_count = self.__serial_data.update()
    for _ in range(line_count):
      nmea_data_tagged = NavilockMd6.__nmea_decoder.decode(self.__serial_data.get())
      if nmea_data_tagged and self._update(nmea_data_tagged, detect_params):
        return True
    return False

  def gga_decode(self, data, detect_params):
    lat = self.convert_lat(data[2][VALUE])
    lon = self.convert_lon(data[4][VALUE])
    #Only use the GPS data when quality index is 1 
    if data[6][VALUE] == '1':
      self.__sensor_data_valid = True
      self.__data['GGA'] = [lat,lon,data[6][VALUE],data[7][VALUE],data[8][VALUE],data[9][VALUE],data[10][VALUE],data[11][VALUE],data[12][VALUE]]
    if detect_params:
      self.__params['GGA'] = [data[2][PARAM],data[4][PARAM],data[6][PARAM],data[7][PARAM],data[8][PARAM],data[9][PARAM],data[10][PARAM],data[11][PARAM],data[12][PARAM]]

  def vtg_decode(self, data, detect_params):
    if self.__sensor_data_valid:
      self.__data['VTG'] = [data[1][VALUE],data[2][VALUE],data[3][VALUE],data[4][VALUE],data[5][VALUE]]
    if detect_params:
      self.__params['VTG'] = [data[1][PARAM],data[2][PARAM],data[3][PARAM],data[4][PARAM],data[5][PARAM]]

  def rmc_decode(self, data, detect_params):
    if self.__sensor_data_valid:
      self.__datetime_update(data[1][VALUE], data[9][VALUE])

  def txt_decode(self, data, detect_params):
    #Do nothing with the TXT data
    pass

  _nmea_senid = {
        "GGA" : gga_decode,
        "VTG" : vtg_decode,
        "RMC" : rmc_decode,
        "TXT" : txt_decode
	}

  def _update(self, data, detect_params):
    try:
      decoder = self._nmea_senid[data[0][VALUE][2:]]
      decoder(self, data, detect_params)
      if len(self.__data) == 2:
        self.data = self.__data['GGA'] + self.__data['VTG']
        self.__sensor_data_valid = False
        self.__data = {}
        return True
      elif len(self.__params) == 2 and detect_params:
        return True
    except KeyError:
      self.__syslog.warn('Unsupported decoded format: {}'.format(data))
    return False

  #Latitude is converted to DD (decimal degrees) notation due to omc045 log format    
  def convert_lat(self, lat):
    if lat:
      lat_degs = parse_int(lat[0:2])
      lat_mins = parse_float(lat[2:])
      lat_hemi = lat[3]
      decimal_degrees = lat_degs + (lat_mins / 60)
      return -decimal_degrees if lat_hemi == 'S' else decimal_degrees

  #Longitude is converted to DD (decimal degrees) notation due to omc045 log format   
  def convert_lon(self, lon):
    if lon:
      lon_degs = parse_int(lon[0:3])
      lon_mins = parse_float(lon[3:])

      if lon_degs and lon_mins:
        lon_hemi = lon[5]
        decimal_degrees = lon_degs + (lon_mins / 60)
        return -decimal_degrees if lon_hemi == 'W' else decimal_degrees

  def __datetime_update(self, utc, date):
    if utc and date:
      hours = (parse_int(utc[0:2])) % 24
      minutes = parse_int(utc[2:4])
      seconds = parse_int(utc[4:6])
      subseconds = parse_int(utc[7:])
      utc_tup = (hours, minutes, seconds, subseconds)
      day = parse_int(date[0:2])
      month = parse_int(date[2:4])
      year = 2000 + parse_int(date[4:])
      date_tup = (year, month, day)
      #calculate the weekday according to the date
      calc_table = [0, 3, 2, 5, 0, 3, 5, 1, 4, 6, 2, 4]
      year = date_tup[0] - (date_tup[1] < 3)
      weekday = ((year + year // 4) - (year // 100) + (year // 400) + calc_table[date_tup[1] - 1] + date_tup[2]) % 7
      self.datetime = (date_tup + (weekday,) + utc_tup)
      if self.__utc_sync and self.__utc_synchronisation_check():
        self.__utc_correct()

  def detect_parameters(self, tag_number = ''):
    if self.read_values(True):
      params = self.__params['GGA'] + self.__params['VTG']
      for param in params:
        self.parameters.append((param[0], param[1], param[2] + tag_number))
      return True

  def flush(self):
    self.__serial_data.flush()

  def __utc_synchronisation_check(self):
    logger_time = omc048.RTC().datetime()
    corrected_gps_time = get_datetime_with_offset(self.datetime, self.__utc_offset[0], self.__utc_offset[1])
    self.datetime = corrected_gps_time
    if abs(obs.date_time_delta(logger_time, corrected_gps_time)) > self.__time_deviation_in_sec:
      return True
    return False

  def __utc_correct(self):
    self.__syslog.info('Logger time deviation to GPS source > {} second'.format(self.__time_deviation_in_sec))
    omc048.RTC().datetime(self.datetime)                    
    self.__syslog.info('Logger time synchronised to GPS source')

